#include <iostream>

#define PNG_DEBUG 3
#include <png.h>
#include <dirent.h>
#include <cstring>


int x, y;

int width, height;
png_byte color_type;
png_byte bit_depth;

png_structp png_ptr;
png_infop info_ptr;
int number_of_passes;
png_bytep * row_pointers;

#define abort_ printf

void read_png_file(char* file_name)
{
	png_const_bytep header = static_cast<png_const_bytep>(malloc(8 * sizeof(char)));    // 8 is the maximum size that can be checked

	/* open file and test for it being a png */
	FILE *fp = fopen(file_name, "rb");
	if (!fp)
		abort_("[read_png_file] File %s could not be opened for reading", file_name);
	fread((void*) header, 1, 8, fp);
	if (png_sig_cmp(header, 0, 8))
		abort_("[read_png_file] File %s is not recognized as a PNG file", file_name);


	/* initialize stuff */
	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

	if (!png_ptr)
		abort_("[read_png_file] png_create_read_struct failed");

	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr)
		abort_("[read_png_file] png_create_info_struct failed");

	if (setjmp(png_jmpbuf(png_ptr)))
		abort_("[read_png_file] Error during init_io");

	png_init_io(png_ptr, fp);
	png_set_sig_bytes(png_ptr, 8);

	png_read_info(png_ptr, info_ptr);

	width = png_get_image_width(png_ptr, info_ptr);
	height = png_get_image_height(png_ptr, info_ptr);
	color_type = png_get_color_type(png_ptr, info_ptr);
	bit_depth = png_get_bit_depth(png_ptr, info_ptr);

	number_of_passes = png_set_interlace_handling(png_ptr);
	png_read_update_info(png_ptr, info_ptr);


	/* read file */
	if (setjmp(png_jmpbuf(png_ptr)))
		abort_("[read_png_file] Error during read_image");

	row_pointers = (png_bytep*) malloc(sizeof(png_bytep) * height);
	for (y=0; y<height; y++)
		row_pointers[y] = (png_byte*) malloc(png_get_rowbytes(png_ptr,info_ptr));

	png_read_image(png_ptr, row_pointers);

	fclose(fp);
}

uint16_t C_RGB(uint8_t r, uint8_t g, uint8_t b){
	return (((r & 0xF8) << 8) | ((g & 0xFC) << 3) | (b >> 3));
}

void output_16b_bmp(char* spriteName, char* path)
{
	if (png_get_color_type(png_ptr, info_ptr) == PNG_COLOR_TYPE_RGB)
		abort_("[output_16b_bmp] input file is PNG_COLOR_TYPE_RGB but must be PNG_COLOR_TYPE_RGBA "
			   "(lacks the alpha channel)");

	if (png_get_color_type(png_ptr, info_ptr) != PNG_COLOR_TYPE_RGBA)
		abort_("[output_16b_bmp] color_type of input file must be PNG_COLOR_TYPE_RGBA (%d) (is %d)",
			   PNG_COLOR_TYPE_RGBA, png_get_color_type(png_ptr, info_ptr));

	FILE* output = fopen(path, "w");
	fprintf(output, "const uint16_t %s[%d * %d] PROGMEM = {\n\t", spriteName, width, height);
	int cursor = -1;

	for (y=0; y<height; y++) {
		png_byte* row = row_pointers[y];
		for (x=0; x<width; x++) {
			cursor++;
			if(cursor == 8){
				fprintf(output, "\n\t");
				cursor = 0;
			}

			png_byte* ptr = &(row[x*4]);

			if(ptr[3] < 255){
				fprintf(output, "0x0120");
			}else{
				fprintf(output, "0x%04x", C_RGB(ptr[0], ptr[1], ptr[2]));
			}

			if((x+1) * (y+1) < width * height){
				fprintf(output, ", ");
			}

			/* set red value to 0 and green value to the blue one */
			ptr[0] = 0;
			ptr[1] = ptr[2];
		}
	}

	if(cursor != 0){
		fprintf(output, "\n");
	}

	fprintf(output, "};\n");
}

void traverseDir(char* inputDir, char* outputDir){
	DIR* dir;
	struct dirent* ent;

	if((dir = opendir(inputDir)) != NULL){
		/* print all the files and directories within directory */
		while((ent = readdir(dir)) != NULL){
			if(!strncmp(ent->d_name, "..", 2) || !strncmp(ent->d_name, ".", 1)) continue;

			char inputPath[120];
			char* spriteName;
			char outputPath[120];
			sprintf(inputPath, "%s/%s", inputDir, ent->d_name);
			spriteName = strtok(ent->d_name, ".");
			sprintf(outputPath, "%s/%s.hpp", outputDir, spriteName);

			read_png_file(inputPath);
			output_16b_bmp(spriteName, outputPath);

			printf("Processed: %s\n", spriteName);
		}
		closedir(dir);
	}else{
		/* could not open directory */
		perror("");
		return;
	}
}

int main(int argc, char* argv[]){
	if(argc != 3){
		printf("Usage: ./BitmapStudio <input dir> <output dir>");
		return 1;
	}

	char* input = argv[1];
	char* output = argv[2];

	traverseDir(input, output);

	//output_16b_bmp(argv[2]);

	return 0;
}
